#pragma once

#include <string_view>

inline constexpr std::string_view AUTOGENERATED_PREFIX_LUA   = R"#(
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --
-- AUTOGENERATED HYPRLAND CONFIG.                        --
-- EDIT THIS CONFIG ACCORDING TO THE WIKI INSTRUCTIONS.  --
-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --

hl.config({ autogenerated = true }) -- remove this line to remove the warning

)#";
inline constexpr char             EXAMPLE_CONFIG_BYTES_LUA[] = {
#embed "../../../example/hyprland.lua"
};

inline constexpr std::string_view EXAMPLE_CONFIG_LUA = {EXAMPLE_CONFIG_BYTES_LUA, sizeof(EXAMPLE_CONFIG_BYTES_LUA)};
