// DO NOT EDIT. Will be overwritten in runtime
// Values here are only for highlighting and static checking
// 1 assumes that a shader either supports this feature or doesn't use any code supporting it
// 0 assumes that shader uses some code supporting the feature but will never have this feature enabled
#define USE_RGBA        1
#define USE_DISCARD     1
#define USE_TINT        1
#define USE_ROUNDING    1
#define USE_CM          1
#define USE_TONEMAP     1
#define USE_ALT_TONEMAP 0
#define USE_SDR_MOD     1
#define USE_BLUR        1
#define USE_ICC         0
#define USE_MIRROR      0
#define USE_BLUR_MATTE  0
